@@ -1,33 +1,46 @@
 import inspect
 import re
+import sys
 import typing
-import warnings
 from collections import OrderedDict
 from decimal import Decimal
 from operator import attrgetter
 from urllib.parse import urljoin
-from uuid import UUID
 
 import uritemplate
 from django.core import validators
 from django.db import models
 from django.utils.encoding import force_str
-from rest_framework import exceptions, permissions, renderers, serializers
+from rest_framework import exceptions, permissions, renderers, serializers, views, viewsets
 from rest_framework.fields import _UnvalidatedField, empty
 from rest_framework.schemas.generators import BaseSchemaGenerator
 from rest_framework.schemas.inspectors import ViewInspector
 from rest_framework.schemas.utils import get_pk_description, is_list_view
 
 from drf_spectacular.app_settings import spectacular_settings
-from drf_spectacular.utils import (
-    TYPE_MAPPING, PolymorphicResponse
-)
+from drf_spectacular.types import OpenApiTypes, resolve_basic_type, PYTHON_TYPE_MAPPING, OPENAPI_TYPE_MAPPING
+from drf_spectacular.utils import PolymorphicResponse
 
 AUTHENTICATION_SCHEMES = {
     cls.authentication_class: cls for cls in spectacular_settings.SCHEMA_AUTHENTICATION_CLASSES
 }
 
 
+def warn(msg):
+    print(f'WARNING: {msg}', file=sys.stderr)
+
+
+def anyisinstance(obj, type_list):
+    return any([isinstance(obj, t) for t in type_list])
+
+
+def force_serializer_instance(serializer):
+    if inspect.isclass(serializer) and issubclass(serializer, serializers.Serializer):
+        return serializer()
+    else:
+        return serializer
+
+
 class ComponentRegistry:
     def __init__(self):
         self.schemas = {}
@@ -55,16 +68,12 @@ class SchemaGenerator(BaseSchemaGenerator):
         """
         view = super().create_view(callback, method, request)
 
-        # circumvent import issues by locally importing
-        from rest_framework.views import APIView
-        from rest_framework.viewsets import GenericViewSet, ViewSet
-
-        if isinstance(view, GenericViewSet) or isinstance(view, ViewSet):
+        if isinstance(view, viewsets.GenericViewSet) or isinstance(view, viewsets.ViewSet):
             action = getattr(view, view.action)
-        elif isinstance(view, APIView):
+        elif isinstance(view, views.APIView):
             action = getattr(view, method.lower())
         else:
-            raise RuntimeError('not supported subclass. Must inherit from APIView')
+            raise RuntimeError('not supported subclass. Must inherit from APIView or subclass of APIView')
 
         if hasattr(action, 'kwargs') and 'schema' in action.kwargs:
             # might already be properly set in case of @action but overwrite for all cases
@@ -82,7 +91,6 @@ class SchemaGenerator(BaseSchemaGenerator):
             return None
 
         # Iterate endpoints generating per method path operations.
-        paths = {}
         for path, method, view in view_endpoints:
             if not self.has_view_permissions(path, method, view):
                 continue
@@ -122,9 +130,6 @@ class SchemaGenerator(BaseSchemaGenerator):
 
 
 class AutoSchema(ViewInspector):
-    request_media_types = []
-    response_media_types = []
-
     method_mapping = {
         'get': 'retrieve',
         'post': 'create',
@@ -139,15 +144,18 @@ class AutoSchema(ViewInspector):
 
         operation['operationId'] = self.get_operation_id(path, method)
         operation['description'] = self.get_description(path, method)
-        operation['parameters'] = sorted(
+
+        parameters = sorted(
             [
                 *self._get_path_parameters(path, method),
                 *self._get_filter_parameters(path, method),
                 *self._get_pagination_parameters(path, method),
                 *self.get_extra_parameters(path, method),
             ],
-            key=lambda p: p.get('name')
+            key=lambda parameter: parameter.get('name')
         )
+        if parameters:
+            operation['parameters'] = parameters
 
         tags = self.get_tags(path, method)
         if tags:
@@ -161,8 +169,6 @@ class AutoSchema(ViewInspector):
         if auth:
             operation['security'] = auth
 
-        self.response_media_types = self.map_renderers(path, method)
-
         operation['responses'] = self._get_response_bodies(path, method)
 
         return operation
@@ -237,37 +243,31 @@ class AutoSchema(ViewInspector):
         """
         model = getattr(getattr(self.view, 'queryset', None), 'model', None)
         parameters = []
+        schema = resolve_basic_type(OpenApiTypes.STR)
+        description = ''
 
         for variable in uritemplate.variables(path):
-            description = ''
-            schema = TYPE_MAPPING[str]
-
-            if model is not None:  # TODO: test this.
-                # Attempt to infer a field description if possible.
+            if not model:
+                warn(f'{self.view.__class__} had no queryset attribute. could not estimate type of parameter "{variable}". defaulting to string.')
+            else:
                 try:
                     model_field = model._meta.get_field(variable)
+                    schema = self._map_model_field(model_field)
+
+                    if model_field.help_text:
+                        description = force_str(model_field.help_text)
+                    elif model_field.primary_key:
+                        description = get_pk_description(model, model_field)
                 except Exception:
-                    model_field = None
+                    pass
 
-                if model_field is not None and model_field.help_text:
-                    description = force_str(model_field.help_text)
-                elif model_field is not None and model_field.primary_key:
-                    description = get_pk_description(model, model_field)
-
-                # TODO are there more relevant PK base classes?
-                if isinstance(model_field, models.IntegerField):
-                    schema = TYPE_MAPPING[int]
-                elif isinstance(model_field, models.UUIDField):
-                    schema = TYPE_MAPPING[UUID]
-
-            parameter = {
+            parameters.append({
                 "name": variable,
                 "in": "path",
                 "required": True,
                 "description": description,
                 'schema': schema,
-            }
-            parameters.append(parameter)
+            })
 
         return parameters
 
@@ -332,30 +332,46 @@ class AutoSchema(ViewInspector):
             mapping['type'] = type
         return mapping
 
-    def _map_field(self, method, field):
-        # Nested Serializers, `many` or not.
+    def _map_model_field(self, field):
+        # in django 3.0 checking both auto and int field is not required but in 2.2 it is
+        if anyisinstance(field, [models.AutoField, models.IntegerField]):
+            return resolve_basic_type(OpenApiTypes.INT)
+        # TODO make this save for django version not having those fields
+        # elif anyisinstance(field, [models.SmallAutoField, models.SmallIntegerField]):
+        #     return resolve_basic_type(OpenApiTypes.INT)
+        # elif anyisinstance(field, [models.BigAutoField, models.BigIntegerField]):
+        #     return resolve_basic_type(OpenApiTypes.INT)
+        elif isinstance(field, models.UUIDField):
+            return resolve_basic_type(OpenApiTypes.UUID)
+        else:
+            warn(f'could not resolve model field "{field}" due to missing mapping')
+            return resolve_basic_type(OpenApiTypes.STR)
+
+    def _map_serializer_field(self, method, field):
         if isinstance(field, serializers.ListSerializer):
             return {
                 'type': 'array',
                 'items': self.resolve_serializer(method, field.child)
             }
+
         if isinstance(field, serializers.Serializer):
-            data = self.resolve_serializer(method, field, nested=True)
-            data['type'] = 'object'
-            return data
+            return self.resolve_serializer(method, field, nested=True)
 
         # Related fields.
         if isinstance(field, serializers.ManyRelatedField):
             return {
                 'type': 'array',
-                'items': self._map_field(method, field.child_relation)
+                'items': self._map_serializer_field(method, field.child_relation)
             }
+
         if isinstance(field, serializers.PrimaryKeyRelatedField):
             model = getattr(field.queryset, 'model', None)
-            if model is not None:
-                model_field = model._meta.pk
-                if isinstance(model_field, models.AutoField):
-                    return {'type': 'integer'}
+
+            if not model:
+                warn(f'unable to map model field {field} due to missing model. defaulting to "string"')
+                return resolve_basic_type(OpenApiTypes.STR)
+            else:
+                return self._map_model_field(model._meta.pk)
 
         # ChoiceFields (single and multiple).
         # Q:
@@ -377,7 +393,7 @@ class AutoSchema(ViewInspector):
             }
             # TODO check this
             if not isinstance(field.child, _UnvalidatedField):
-                map_field = self._map_field(method, field.child)
+                map_field = self._map_serializer_field(method, field.child)
                 items = {
                     "type": map_field.get('type')
                 }
@@ -388,45 +404,27 @@ class AutoSchema(ViewInspector):
 
         # DateField and DateTimeField type is string
         if isinstance(field, serializers.DateField):
-            return {
-                'type': 'string',
-                'format': 'date',
-            }
+            return resolve_basic_type(OpenApiTypes.DATE)
 
         if isinstance(field, serializers.DateTimeField):
-            return {
-                'type': 'string',
-                'format': 'date-time',
-            }
+            return resolve_basic_type(OpenApiTypes.DATETIME)
 
-        # "Formats such as "email", "uuid", and so on, MAY be used even though undefined by this specification."
-        # see: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.2.md#data-types
-        # see also: https://swagger.io/docs/specification/data-models/data-types/#string
         if isinstance(field, serializers.EmailField):
-            return {
-                'type': 'string',
-                'format': 'email'
-            }
+            return resolve_basic_type(OpenApiTypes.EMAIL)
 
         if isinstance(field, serializers.URLField):
-            return {
-                'type': 'string',
-                'format': 'uri'
-            }
+            return resolve_basic_type(OpenApiTypes.URI)
 
         if isinstance(field, serializers.UUIDField):
-            return {
-                'type': 'string',
-                'format': 'uuid'
-            }
+            return resolve_basic_type(OpenApiTypes.UUID)
 
         if isinstance(field, serializers.IPAddressField):
-            content = {
-                'type': 'string',
-            }
-            if field.protocol != 'both':
-                content['format'] = field.protocol
-            return content
+            if 'IPv4' == field.protocol:
+                return resolve_basic_type(OpenApiTypes.IP4)
+            elif 'IPv6' == field.protocol:
+                return resolve_basic_type(OpenApiTypes.IP6)
+            else:
+                return resolve_basic_type(OpenApiTypes.STR)
 
         # DecimalField has multipleOf based on decimal_places
         if isinstance(field, serializers.DecimalField):
@@ -442,16 +440,12 @@ class AutoSchema(ViewInspector):
             return content
 
         if isinstance(field, serializers.FloatField):
-            content = {
-                'type': 'number'
-            }
+            content = resolve_basic_type(OpenApiTypes.FLOAT)
             self._map_min_max(field, content)
             return content
 
         if isinstance(field, serializers.IntegerField):
-            content = {
-                'type': 'integer'
-            }
+            content = resolve_basic_type(OpenApiTypes.INT)
             self._map_min_max(field, content)
             # 2147483647 is max for int32_size, so we use int64 for format
             if int(content.get('maximum', 0)) > 2147483647 or int(content.get('minimum', 0)) > 2147483647:
@@ -459,23 +453,28 @@ class AutoSchema(ViewInspector):
             return content
 
         if isinstance(field, serializers.FileField):
-            return {
-                'type': 'string',
-                'format': 'binary'
-            }
+            return resolve_basic_type(OpenApiTypes.BINARY)
 
         if isinstance(field, serializers.SerializerMethodField):
             method = getattr(field.parent, field.method_name)
             return self._map_type_hint(method)
 
-        # Simplest cases, default to 'string' type:
-        FIELD_CLASS_SCHEMA_TYPE = {
-            serializers.BooleanField: 'boolean',
-            serializers.JSONField: 'object',
-            serializers.DictField: 'object',
-            serializers.HStoreField: 'object',
-        }
-        return {'type': FIELD_CLASS_SCHEMA_TYPE.get(field.__class__, 'string')}
+        if isinstance(field, serializers.BooleanField):
+            return resolve_basic_type(OpenApiTypes.BOOL)
+
+        if anyisinstance(field, [serializers.JSONField, serializers.DictField, serializers.HStoreField]):
+            return resolve_basic_type(OpenApiTypes.OBJECT)
+
+        if isinstance(field, serializers.CharField):
+            return resolve_basic_type(OpenApiTypes.STR)
+
+        # TODO serializer fields
+        # serializers.ReadOnlyField
+        # serializers.CreateOnlyDefault
+        # serializers.RegexField
+
+        warn(f'could not resolve serializer field {field}. defaulting to "string"')
+        return resolve_basic_type(OpenApiTypes.STR)
 
     def _map_min_max(self, field, content):
         if field.max_value:
@@ -494,7 +493,7 @@ class AutoSchema(ViewInspector):
             if field.required:
                 required.append(field.field_name)
 
-            schema = self._map_field(method, field)
+            schema = self._map_serializer_field(method, field)
             if field.read_only:
                 schema['readOnly'] = True
             if field.write_only:
@@ -505,6 +504,7 @@ class AutoSchema(ViewInspector):
                 schema['default'] = field.default
             if field.help_text:
                 schema['description'] = str(field.help_text)
+
             self._map_field_validators(field, schema)
 
             properties[field.field_name] = schema
@@ -521,7 +521,7 @@ class AutoSchema(ViewInspector):
     def _map_field_validators(self, field, schema):
         for v in field.validators:
             # "Formats such as "email", "uuid", and so on, MAY be used even though undefined by this specification."
-            # https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.2.md#data-types
+            # https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.3.md#data-types
             if isinstance(v, validators.EmailValidator):
                 schema['format'] = 'email'
             if isinstance(v, validators.URLValidator):
@@ -556,25 +556,19 @@ class AutoSchema(ViewInspector):
         if not hint:
             hint = typing.get_type_hints(method).get('return')
 
-        if hint in TYPE_MAPPING:
-            return TYPE_MAPPING[hint]
-        elif hint.__origin__ is typing.Union:
-            sub_hints = [
-                self._map_type_hint(method, sub_hint)
-                for sub_hint in hint.__args__ if sub_hint is not type(None)  # noqa
-            ]
-            if type(None) in hint.__args__ and len(sub_hints) == 1:
-                return {**sub_hints[0], 'nullable': True}
-            elif type(None) in hint.__args__:
-                return {'oneOf': [{**sub_hint, 'nullable': True} for sub_hint in sub_hints]}
+        if hasattr(method, '_spectacular_annotation'):
+            if isinstance(force_serializer_instance(method._spectacular_annotation), serializers.BaseSerializer):
+                return self._map_serializer_field(method, force_serializer_instance(method._spectacular_annotation))
+            elif method._spectacular_annotation in OPENAPI_TYPE_MAPPING:
+                return resolve_basic_type(method._spectacular_annotation)
             else:
-                return {'oneOf': sub_hints}
+                warn(f'unknown decorated type "{repr(method._spectacular_annotation)}". defaulting to string')
+                return resolve_basic_type(OpenApiTypes.STR)
+        elif hint in PYTHON_TYPE_MAPPING:
+            return resolve_basic_type(hint)
         else:
-            warnings.warn(
-                'type hint for SerializerMethodField function "{}" is unknown. '
-                'defaulting to string.'.format(method.__name__)
-            )
-            return {'type': 'string'}
+            warn(f'type hint for SerializerMethodField function "{method.__name__}" is unknown. defaulting to string.')
+            return resolve_basic_type(OpenApiTypes.STR)
 
     def _get_paginator(self):
         pagination_class = getattr(self.view, 'pagination_class', None)
@@ -603,7 +597,7 @@ class AutoSchema(ViewInspector):
         try:
             return view.get_serializer()
         except exceptions.APIException:
-            warnings.warn(
+            warn(
                 '{}.get_serializer() raised an exception during '
                 'schema generation. Serializer fields will not be '
                 'generated for {} {}.'.format(view.__class__.__name__, method, path)
@@ -614,16 +608,14 @@ class AutoSchema(ViewInspector):
         if method not in ('PUT', 'PATCH', 'POST'):
             return {}
 
-        request_media_types = self.map_parsers(path, method)
-
-        serializer = self._get_serializer(path, method)
+        serializer = force_serializer_instance(self.get_request_serializer(path, method))
 
         if isinstance(serializer, serializers.Serializer):
             schema = self.resolve_serializer(method, serializer)
         else:
-            warnings.warn(
-                'could not resolve request body for {} {}. defaulting to generic '
-                'free-form object. (maybe annotate a Serializer class?)'.format(method, path)
+            warn(
+                f'could not resolve request body for {method} {path}. defaulting to generic '
+                'free-form object. (maybe annotate a Serializer class?)'
             )
             schema = {
                 'type': 'object',
@@ -636,13 +628,15 @@ class AutoSchema(ViewInspector):
             return {}
 
         return {
-            'content': {mt: {'schema': schema} for mt in request_media_types}
+            'content': {
+                request_media_types: {'schema': schema} for request_media_types in self.map_parsers(path, method)
+            }
         }
 
     def _get_response_bodies(self, path, method):
         response_serializers = self.get_response_serializers(path, method)
 
-        if isinstance(response_serializers, serializers.Serializer) or isinstance(response_serializers, PolymorphicResponse):
+        if anyisinstance(force_serializer_instance(response_serializers), [serializers.Serializer, PolymorphicResponse]):
             if method == 'DELETE':
                 return {'204': {'description': 'No response body'}}
             return {'200': self._get_response_for_code(path, method, response_serializers)}
@@ -653,32 +647,28 @@ class AutoSchema(ViewInspector):
                 for code, serializer in response_serializers.items()
             }
         else:
-            warnings.warn(
-                'could not resolve response for {} {}. defaulting '
-                'to generic free-form object.'.format(method, path)
-            )
+            warn(f'could not resolve response for {method} {path}. defaulting to generic free-form object.')
             schema = {
                 'type': 'object',
                 'description': 'Unspecified response body',
             }
             return {'200': self._get_response_for_code(path, method, schema)}
 
-    def _get_response_for_code(self, path, method, serializer_instance):
+    def _get_response_for_code(self, path, method, serializer):
         # convenience feature: auto instantiate serializer classes
-        if inspect.isclass(serializer_instance) and issubclass(serializer_instance, serializers.Serializer):
-            serializer_instance = serializer_instance()
+        serializer = force_serializer_instance(serializer)
 
-        if not serializer_instance:
+        if not serializer:
             return {'description': 'No response body'}
-        elif isinstance(serializer_instance, serializers.Serializer):
-            schema = self.resolve_serializer(method, serializer_instance)
+        elif isinstance(serializer, serializers.Serializer):
+            schema = self.resolve_serializer(method, serializer)
             if not schema:
                 return {'description': 'No response body'}
-        elif isinstance(serializer_instance, PolymorphicResponse):
+        elif isinstance(serializer, PolymorphicResponse):
             # custom handling for @extend_schema's injection of polymorphic responses
             schemas = []
 
-            for serializer in serializer_instance.serializers:
+            for serializer in serializer.serializers:
                 assert isinstance(serializer, serializers.Serializer)
                 schema_option = self.resolve_serializer(method, serializer)
                 if schema_option:
@@ -687,18 +677,18 @@ class AutoSchema(ViewInspector):
             schema = {
                 'oneOf': schemas,
                 'discriminator': {
-                    'propertyName': serializer_instance.resource_type_field_name
+                    'propertyName': serializer.resource_type_field_name
                 }
             }
-        elif isinstance(serializer_instance, serializers.ListSerializer):
-            schema = self.resolve_serializer(method, serializer_instance.child)
-        elif isinstance(serializer_instance, dict):
+        elif isinstance(serializer, serializers.ListSerializer):
+            schema = self.resolve_serializer(method, serializer.child)
+        elif isinstance(serializer, dict):
             # bypass processing and use given schema directly
-            schema = serializer_instance
+            schema = serializer
         else:
             raise ValueError('Serializer type unsupported')
 
-        if isinstance(serializer_instance, serializers.ListSerializer) or is_list_view(path, method, self.view):
+        if isinstance(serializer, serializers.ListSerializer) or is_list_view(path, method, self.view):
             # TODO i fear is_list_view is not covering all the cases
             schema = {
                 'type': 'array',
@@ -710,11 +700,11 @@ class AutoSchema(ViewInspector):
 
         return {
             'content': {
-                mt: {'schema': schema} for mt in self.response_media_types
+                mt: {'schema': schema} for mt in self.map_renderers(path, method)
             },
             # Description is required by spec, but descriptions for each response code don't really
             # fit into our model. Description is therefore put into the higher level slots.
-            # https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.2.md#responseObject
+            # https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.3.md#responseObject
             'description': ''
         }
 
@@ -746,13 +736,13 @@ class AutoSchema(ViewInspector):
             # add placeholder to prevent recursion loop
             self.registry.schemas[name] = None
 
-            mapped = self._map_serializer(method, serializer, nested)
+            schema = self._map_serializer(method, serializer, nested)
             # empty serializer - usually a transactional serializer.
             # no need to put it explicitly in the spec
-            if not mapped['properties']:
+            if not schema['properties']:
                 del self.registry.schemas[name]
                 return {}
             else:
-                self.registry.schemas[name] = mapped
+                self.registry.schemas[name] = schema
 
         return {'$ref': '#/components/schemas/{}'.format(name)}
