@@ -1,26 +1,8 @@
 import inspect
-import warnings
-from datetime import date, datetime
-from decimal import Decimal
-from uuid import UUID
 
 from rest_framework.settings import api_settings
 
-VALID_TYPES = ['integer', 'number', 'string', 'boolean']
-
-TYPE_MAPPING = {
-    float: {'type': 'number', 'format': 'float'},
-    bool: {'type': 'boolean'},
-    str: {'type': 'string'},
-    bytes: {'type': 'string', 'format': 'binary'},  # or byte?
-    int: {'type': 'integer'},
-    UUID: {'type': 'string', 'format': 'uuid'},
-    Decimal: {'type': 'number', 'format': 'double'},
-    datetime: {'type': 'string', 'format': 'date-time'},
-    date: {'type': 'string', 'format': 'date'},
-    None: {},
-    type(None): {},
-}
+from drf_spectacular.types import resolve_basic_type
 
 
 class PolymorphicResponse:
@@ -31,27 +13,33 @@ class PolymorphicResponse:
 
 class OpenApiSchemaBase:
     """ reusable base class for objects that can be translated to a schema """
+
     def to_schema(self):
         raise NotImplementedError('translation to schema required.')
 
 
-class QueryParameter(OpenApiSchemaBase):
-    def __init__(self, name, description='', required=False, type=str):
+class ExtraParameter(OpenApiSchemaBase):
+    QUERY = 'query'
+    PATH = 'path'
+    HEADER = 'header'
+
+    def __init__(self, name, type=str, location=QUERY, required=False, description=''):
         self.name = name
-        self.description = description
-        self.required = required
         self.type = type
+        self.location = location
+        self.required = required
+        self.description = description
 
     def to_schema(self):
-        if self.type not in TYPE_MAPPING:
-            warnings.warn('{} not a mappable type'.format(self.type))
-        return {
+        schema = {
+            'in': self.location,
             'name': self.name,
-            'in': 'query',
+            'schema': resolve_basic_type(self.type),
             'description': self.description,
-            'required': self.required,
-            'schema': TYPE_MAPPING.get(self.type)
         }
+        if self.location != self.PATH:
+            schema['required'] = self.required
+        return schema
 
 
 def extend_schema(
@@ -135,3 +123,17 @@ def extend_schema(
             return f
 
     return decorator
+
+
+def extend_schema_field(response):
+    """
+    decorator for SerializerMethodField(). always takes precedence over a type hint (if one was given).
+
+    :param response: accepts a Serializer or OpenApiTypes
+    :return:
+    """
+    def decorator(f):
+        f._spectacular_annotation = response
+        return f
+
+    return decorator
